import pandas as pd

def create_gene_list_overlap_df(term_gene_maps, parent_offspring_map):

    for ontology in term_gene_maps:
        term_gene_map = term_gene_maps[ontology]
        overlap_data = get_overlap_data_for_terms(term_gene_map, parent_offspring_map)

        df = pd.DataFrame.from_dict(overlap_data)

    return df


def get_overlap_data_for_terms(term_gene_map, parent_offspring_map):

    overlap_data = []

    for parent, offspring in parent_offspring_map.items():

        genes_related_to_offspring = get_all_genes_related_to_offspring(parent, offspring, term_gene_map)

        overlap_data_for_parent = get_overlap_data(genes_a, genes_b, genes_related_to_offspring, parent)
        overlap_data.append(overlap_data_for_parent)

    return overlap_data


def get_overlap_data(genes_a, genes_b, genes_related_to_offspring, parent):

    a_only, b_only, a_intersect_b = compute_input_genes_overlap(genes_a, genes_b)
    a_intersect_go, b_intersect_go, a_intersect_b_intersect_go = compute_input_genes_ontology_overlap(a_only, b_only,
                                                                                                      a_intersect_b,
                                                                                                      genes_related_to_offspring)

    a_only, b_only, a_intersect_b = len(a_only), len(b_only), len(a_intersect_b)
    a_intersect_go, b_intersect_go, a_intersect_b_intersect_go = len(a_intersect_go), len(b_intersect_go), len(a_intersect_b_intersect_go)
    genes_related_to_offspring = len(genes_related_to_offspring)

    return {"term": parent,
            "a_b": a_intersect_b,
            "a": a_only,
            "b": b_only,
            "a_b_go": a_intersect_b_intersect_go,
            "a_go": a_intersect_go,
            "b_go": b_intersect_go,
            "go": genes_related_to_offspring}

def get_all_genes_related_to_offspring(parent, offspring, term_gene_map):

    all_genes_related_to_offspring_terms = set()

    terms_found = [term for term in offspring if term in term_gene_map]
    for term in terms_found:

        genes_related_to_term = term_gene_map[term]
        all_genes_related_to_offspring_terms.update(genes_related_to_term)

    all_genes_related_to_offspring_terms.add(parent)

    return all_genes_related_to_offspring_terms



import pandas as pd

from biomartian.bm_queries.bm_query import get_bm
from kegg.lib import get_kegg
from joblib import Memory

MEMORY = Memory(cachedir="./")


def terms_to_genes(ontology, dataset, all_terms):

    """Get the genes belonging to each ontology term."""

    all_terms = pd.Series(all_terms)
    ontology_to_biomart_id = {"CC": "go_id", "MF": "go_id", "BP": "go_id",
                               "REACTOME": "reactome", "KEGG": "kegg"}
    ontology_id = ontology_to_biomart_id[ontology]

    map_df = get_term_to_gene_mapping(ontology_id, dataset)

    requested_genes_only = map_df[map_df["Term"].isin(all_terms)]

    return requested_genes_only


def set_up_r_session_and_get_mart(dataset):

    r_session = R()
    r_session.load_library("biomaRt")

    get_mart_command = 'mart <- useMart("ensembl", dataset="{}")'.format(dataset)
    r_session(get_mart_command)

    return r_session


@MEMORY.cache(verbose=0)
def get_term_to_gene_mapping(ontology_id, dataset):
    """"Get 2 col df with ontology term and related genes."""
    if ontology_id == "kegg":
        return get_term_to_gene_mapping_kegg(dataset)
    elif ontology_id == "go_id":
        go_id = get_bm("external_gene_name", ontology_id, dataset, "ensembl")
        go_id.columns = ["Gene", "Term"]
        go_id = go_id.dropna()
        return go_id


def get_term_to_gene_mapping_kegg(species):

    species_to_kegg_map = {"rnorvegicus_gene_ensembl": "rno",
                           "hsapiens_gene_ensembl": "hsa",
                           "mmusculus_gene_ensembl": "mmu"}

    species = species_to_kegg_map[species]

    add_definitions = False
    kegg_df = get_kegg(species, add_definitions)

    kegg_df = kegg_df[["kegg_pathway", "gene"]]
    kegg_df.columns = ["Term", "Gene"]

    return kegg_df



from scipy.stats import hypergeom, rankdata, chisqprob, fisher_exact
import logging

from joblib import Memory

MEMORY = Memory(cachedir="./")


@MEMORY.cache(verbose=0)
def compute_statistics(df, gene_vectors):

    """Compute statistics for the gene lists overlap counts."""

    logging.info("Performing hypergeometric tests.")
    df = compute_go_overrepresentation_analysis(df)
    df = compute_odds_ratio(df)

    stat_method = find_appropriate_statistical_test(gene_vectors)

    if stat_method == "LAP":

        logging.info("Computing the Leisering, Alonzo and Pepe test.")
        df = lap_ppv(df)
        fdr_cols = ["a_hpgm_go", "b_hpgm_go", "chi_sq"]

    elif stat_method == "FE":

        logging.info("Computing the Fisher Exact test.")
        df = fishers_exact(df)
        fdr_cols = ["a_hpgm_go", "b_hpgm_go", "fisher"]

    elif stat_method == "NA":

        fdr_cols = ["a_hpgm_go"]

    logging.info("Computing FDR.")
    df = compute_fdr(df, fdr_cols)

    return df


def compute_odds_ratio(df):

    df["a_odds"] = (df["a_I_go"] / df["a"]) / \
                   ((df["U"] - (df["a"] + df["go_subtree"] - df["a_I_go"])) / (df["U"] - df["a"]))


    if "b" in df:
        df["b_odds"] = (df["b_I_go"] / df["b"]) / \
                    ((df["U"] - (df["b"] + df["go_subtree"] - df["b_I_go"])) / (df["U"] - df["b"]))
    return df


def fishers_exact(df):

    """Compute Fisher Exact for diff in a and b go overlap."""

    df["fisher"] = df.apply(lambda row:
                            fisher_exact([
                                [row["a_I_go"], row["a"] - row["a_I_go"]],
                                [row["b_I_go"], row["b"] - row["b_I_go"]]
                            ], alternative="two-sided")[1],
                            axis=1)

    return df


@MEMORY.cache(verbose=0)
def compute_go_overrepresentation_analysis(df):

    """Compute p-val for gene list representation in GO category.

    Uses hypergeometric test, supposed to be similar to GOstats.
    """

    df["a_hpgm_go"] = df.apply(lambda row: hypergeom.sf(row["a_I_go"] - 1,
                                                 row["U"],
                                                 row["go_subtree"],
                                                 row["a"]), axis=1)

    if "b" in df:
        df["b_hpgm_go"] = df.apply(lambda row: hypergeom.sf(row["b_I_go"] - 1,
                                                            row["U"],
                                                            row["go_subtree"],
                                                            row["b"]), axis=1)

    return df


def compute_fdr(df, columns):
    """Computes the false discovery rate for a list of columns."""

    for column in columns:
        cp = df[column]
        fdr_col = column + "_fdr"

        # Using the length of the vector without NAs because NAs indicate
        # that an experiment wasn't performed so you do not have to adjust
        # for it.
        df[fdr_col] = cp * len(cp.dropna()) / rankdata(cp)

        df.loc[df[fdr_col] > 1, fdr_col] = 1

    return df


def lap_ppv(df):

    """Computes the Generalized Score Statistic for Comparison of PPVs

    (The generalized score statistic was proposed by Leisenring, Alonzo and Pepe
    (2000), hence the name lap_ppv.)

    This is a test used to compare the positive predictive value (PPV) of
    two gene lists with regards to the gene list associated with an ontology.

    That is, for three gene lists A and B and O (the genes associated
    with an ontology term), it computes the PPV for A and B with regards to O,
    then compares the two PPVs to see whether they are significantly different.

    For a gene list A and an ontology gene list O, the PPV is just

    PPV = TP/(TP + FP)

    where

    TP = A set subtract O

    while

    FP = A set intersect O.

    This method should be used when lists A and B are not disjoint and one list
    is not a proper subset of the other.

    Alternative description:

    'Performs a test for differences in (positive and negative) predictive
    values of two binary diagnostic tests using a generalized score statistic
    proposed by Leisenring, Alonzo and Pepe'
    ( From https://cran.r-project.org/web/packages/DTComPair/DTComPair.pdf )

    Literature:

    Leisenring, W., Alonzo, T., and Pepe, M. S. (2000). Comparisons of
    predictive values of binary medical diagnostic tests for paired designs.
    Biometrics, 56(2):345-51.

    Clara-Cecilie Gunther, Mette Langaas and Stian Lydersen, 2006: Statistical
    Hypothesis Testing of Association Between Two Lists of Genes for a Given
    Gene Class

    Beisvaag et al 2006: GeneTools - application for functional annotation
    and statistical hypothesis testing.
    (The Leisering test is described there as "the intersecting target-target
    situation" and used in the exactly same way as here.)

    ------------------------------------------------------------------------
    This is the R code this function is based on (thanks Mette Langaas!)

    LAPtestPPV <- function(datamat)
    {
    #faar inn data paa formen n1,n2,n3,...,n8
    x1 <- datamat[,5]
    n1 <- datamat[,1]+datamat[,5]
    x2 <- datamat[,6]
    n2 <- datamat[,2]+datamat[,6]
    x3 <- datamat[,7]
    n3 <- datamat[,3]+datamat[,7]

    nA <- n1+n2
    nB <- n1+n3

    xAG <- x1+x2
    xBG <- x1+x3

    Dstrek <- (xAG+xBG)/(nA+nB)
    Zstrek <- nB/(nA+nB)

    teller2 <- (xBG-nB*(xAG+xBG)/(nA+nB))^2
    # 6 sums: first 3 with G*: AsnittB, AsnittB*, A*snittB, then with G
    nevner2 <-
((0-Dstrek)^2)*((n1-x1)*(1-2*Zstrek)^2+(n2-x2)*(0-1*Zstrek)^2+(n3-x3)*(1-1*Zstrek)^2)+((1-Dstrek)^2)*(x1*(1-2*Zstrek)^2+x2*(0-1*Zstrek)^2+x3*(1-1*Zstrek)^2)


    testobs <- teller2/nevner2
    pvalchisq <- pchisq(testobs,1,lower.tail=F)
    return(list(pval=pvalchisq,testobs=testobs))

    }

    """

    x1 = df["a_I_b"]
    n1 = x1 + df["a_I_b_I_go"]
    x2 = df["a_M_b_I_go"]
    n2 = x2 + df["a_I_b_I_go"]
    x3 = df["b_M_a_I_go"]
    n3 = df["a_I_b_I_go"] + x3

    nA = n1 + n2
    nB = n1 + n3

    xAG = x1 + x2
    xBG = x1 + x3

    d_prime = (xAG + xBG) / (nA + nB)
    z_prime = nB / (nA + nB)

    numerator = (xBG - nB * (xAG + xBG) / (nA + nB))**2

    denominator = ((0 - d_prime)**2) * \
                  ((n1 - x1) * (1 - 2 * z_prime)**2 + (n2 - x2)* \
                   (0 - 1 * z_prime)**2+(n3 - x3)*(1 - 1 * z_prime)**2)+ \
                   ((1 - d_prime)**2) * \
                   (x1 * (1 - 2 * z_prime)**2 + x2 * (0 - 1 * z_prime)**2 + \
                    x3 * (1 - 1 * z_prime)**2)

    test_observation = numerator / denominator

    df["chi_sq"] = chisqprob(test_observation, 1)
    df["test_obs"] = test_observation

    return df


def find_appropriate_statistical_test(gene_vectors):

    """Find the appropriate statistical test.

    The statistical test to use to decide whether there is a statistically
    significant difference between the overrepresentation of 1) gene list a in GO
    and 2) gene list b in GO.

    If there is only one input gene list, or the lists contain the same genes,
    do not perform a significance test of the difference.

    If one input gene list is a subset of the other, or they are disjoint,
    use Fisher's Exact.

    If the lists are intersecting, but they are not proper subsets/supersets of
    each other, use the LAP test.
    """

    if len(gene_vectors) == 1:
        return "NA"

    genes_a, genes_b = gene_vectors
    genes_a, genes_b = set(genes_a), set(genes_b)

    if genes_a == genes_b:
        logging.warning("Gene list A is equal to B. Only hypergeometric tests"
                        " are performed.")
        return "NA"
    elif genes_a.issubset(genes_b):
        logging.warning("Gene list A is a proper subset of list B."
                        " The LAP test is used even though Fisher's Exact"
                        " would give better P-values.")
        return "LAP"
    elif genes_b.issubset(genes_a):
        logging.warning("Gene list B is a proper subset of list A."
                        " The LAP test is used even though Fisher's Exact"
                        " would give better P-values.")
        return "LAP"
    elif not genes_a.intersection(genes_b):
        return "FE"
    else:
        # The gene lists are intersecting, but not proper
        # subsets so Fishers Exact can't be used.
        return "LAP"




from joblib import Memory
from kegg.lib import get_pathway_to_definition_map
import godb

MEMORY = Memory(cachedir="./")


def get_and_insert_annotations(df, dataset, nb_cpus):

    """Get the definition of each ontology category.

    Gives each ontology category a human readable
    definition.

    Example (term and definition):
    GO:0005706 A thread-like connection joining two regions of ectopically paired
               polytene chromosomes."""

    annotations = get_annotations(df["go_root"], dataset, nb_cpus)

    annotations = annotations.reset_index(drop=True)
    annotations.columns = ["go_root", "go_annotation", "ontology"]

    df = df.reset_index(drop=True)

    return df.merge(annotations, on="go_root")


@MEMORY.cache(verbose=0)
def get_annotations(terms, dataset, nb_cpus):
    """Get annotations for each term in terms."""

    kegg_terms = terms[terms.apply(lambda term: "GO" not in term)]

    go_terms = terms[terms.apply(lambda term: "GO" in term)]

    assert len(kegg_terms) + len(go_terms) == len(terms), \
        "The sum of KEGG and GO terms is not equal to the total number of terms."

    go_annotations = get_annotations_for_go(go_terms)
    kegg_annotations = get_annotations_for_kegg(kegg_terms, dataset)

    annotations = go_annotations.append(kegg_annotations).sort()
    annotations.columns = [0, 1, 2]

    return annotations


@MEMORY.cache(verbose=0)
def get_annotations_for_go(terms):

    """Get the definition of GO terms."""

    df = godb.get_annotations()

    df = df[["GO id", "Term", "Ontology"]]

    df = df[df["GO id"].isin(terms)]

    df.columns = ["go_root", "go_annotation", "ontology"]

    return df.drop_duplicates()


@MEMORY.cache(verbose=0)
def get_annotations_for_kegg(terms, dataset):

    """Get KEGG term definitions from kg Python package."""

    species_to_kegg_map = {"rnorvegicus_gene_ensembl": "rno",
                           "hsapiens_gene_ensembl": "hsa",
                           "mmusculus_gene_ensembl": "mmu"}
    species = species_to_kegg_map[dataset]
    kegg_df = get_pathway_to_definition_map(species)

    kegg_df = kegg_df[["kegg_pathway", "kegg_pathway_definition"]]
    kegg_df = kegg_df[kegg_df["kegg_pathway"].isin(terms)]
    kegg_df["ontology"] = "KG"

    kegg_df.columns = ["go_root", "go_annotation", "ontology"]

    return kegg_df.drop_duplicates()



import sys

import pandas as pd

from godb import get_offspring
from kegg.lib import get_pathway_to_definition_map

from joblib import Memory

MEMORY = Memory(cachedir="./")


@MEMORY.cache(verbose=0)
def create_parent_offspring_maps(ontologies, dataset):

    parent_offspring_maps = []
    for ontology in ontologies:
        parent_offspring_map = create_parent_offspring_map(ontology, dataset)
        parent_offspring_maps.append(parent_offspring_map)

    return pd.concat(parent_offspring_maps, axis=1)


def create_parent_offspring_map(ontology, dataset):

    """Create 2 col long format df where col 1 parent col 2 offspring."""

    if ontology in ["CC", "BP", "MF"]:
        return create_parent_offspring_map_go(ontology)
    elif ontology == "KEGG":
        return create_parent_offspring_map_kegg(dataset)
    else:
        print("Ontology {} not found.".format(ontology))
        sys.exit(1)


@MEMORY.cache(verbose=0)
def create_parent_offspring_map_kegg(species):

    """Get parent offspring map kegg.

    Since KEGG non-hierarchical this is merely a df where the KEGG term in the
    first and second column are the same."""

    # kegg uses a three letter id to denote species; hsa instead of hsapiens,
    # mmu instead of mmusculus etc.
    kegg_id = species[0:3]
    kegg_df = get_pathway_to_definition_map(kegg_id)

    # kegg is not hierarchical so the parent offspring map is just
    # the parent-parent map
    kegg_df = kegg_df[["kegg_pathway", "kegg_pathway"]]
    kegg_df.columns = ["Parent", "Offspring"]
    return kegg_df


@MEMORY.cache(verbose=0)
def create_parent_offspring_map_go(ontology):

    """Get parent offspring map for go terms."""

    df = get_offspring(ontology, relations=["is_a", "part_of"])

    all_terms = pd.Series(list(set(df.ix[:, 0]) | set(df.ix[:, 1])))
    parent_parent = pd.concat([all_terms, all_terms], axis=1)

    cols = ["Offspring", "Parent"]
    df.columns, parent_parent.columns = cols, cols

    return pd.concat([df, parent_parent])



try:
    from functools import lru_cache
except ImportError:
    from repoze.lru import lru_cache


@lru_cache(maxsize=32)
def compute_input_genes_overlap(genes_a, genes_b):

    genes_a, genes_b = set(genes_a), set(genes_b)
    a_only = genes_a - genes_b
    b_only = genes_b - genes_a
    a_intersect_b = (genes_a.intersection(genes_b) - a_only) - b_only

    return a_only, b_only, a_intersect_b


def compute_input_genes_ontology_overlap(a_only, b_only, a_and_b, ontology):

    a_and_ontology = a_only.intersection(ontology)
    b_and_ontology = b_only.intersection(ontology)
    a_and_b_and_ontology = a_and_b.intersection(ontology)

    return a_and_ontology, b_and_ontology, a_and_b_and_ontology




from go_overlap.compute_set_overlaps.set_overlaps import \
    compute_non_go_overlaps


def prepare_gene_vectors_for_computation(gene_vectors, all_genes_vector):

    """Turn gene vector(s) into bool lists and compute intersections/diffs.

    First turn vector(s) into bool lists, then, if there are several,
    compute their differences and intersection."""

    bool_vectors = []
    for gene_vector in gene_vectors:
        bool_vector = turn_input_gene_lists_into_bool_vectors(gene_vector,
                                                              all_genes_vector)
        bool_vectors.append(bool_vector)

    return compute_non_go_overlaps(bool_vectors)




def turn_input_gene_lists_into_bool_vectors(gene_vector, all_genes_vector):

    """Turn lists of genes into vectors of bools.

    Turns all gene names into uppercase before comparison to ensure case of gene
    names do not matter."""

    all_genes_vector = all_genes_vector.str.upper()

    gene_vector = gene_vector.str.upper()
    bool_vector = all_genes_vector.isin(gene_vector)

    return bool_vector



import logging
import sys
import pandas as pd
from numpy import uint8, logical_or, array_split
import multiprocessing

from go_overlap.gene_to_terms.terms_to_genes import terms_to_genes
from go_overlap.go_hierarchy.compute_parent_offspring_maps import \
    create_parent_offspring_map

from joblib import Memory


MEMORY = Memory(cachedir="./")

@MEMORY.cache(verbose=0)
def create_gene_subtree_matrix(ontologies, dataset, all_genes_vector, nb_cpus):

    """Create df showing which genes are in which term subtree."""

    non_hierarchical_ontologies = ["KEGG"]

    subtree_to_genes_dfs = []
    for ontology in ontologies:

        parent_offspring_map = create_parent_offspring_map(ontology, dataset)

        all_terms = parent_offspring_map["Offspring"].drop_duplicates()
        terms_to_genes_df = terms_to_genes(ontology, dataset, all_terms)

        nb_terms = len(terms_to_genes_df["Term"].drop_duplicates())
        logging.info("Make term gene map into bool array for the {nb_terms} " \
                     "terms in {ontology} for which associated genes can be " \
                     "found." .format(**vars()))

        terms_to_genes_bool_df = create_gene_list_bool_array(all_genes_vector,
                                                             terms_to_genes_df)
        # terms_to_genes_bool_df["ontology"] = ontology
        if ontology in non_hierarchical_ontologies:
            subtree_to_genes_df = terms_to_genes_bool_df
        else:
            logging.info("Turn term gene bool map into term subtree gene " \
                         "bool map.")
            subtree_to_genes_df = get_all_genes_in_term_subtree(
                parent_offspring_map, terms_to_genes_bool_df, nb_cpus)

        subtree_to_genes_dfs.append(subtree_to_genes_df)

    if len(subtree_to_genes_dfs) > 1:
        return pd.concat(subtree_to_genes_dfs, axis=1)
    elif len(subtree_to_genes_dfs) == 1:
        return subtree_to_genes_dfs[0]


@MEMORY.cache(verbose=0)
def create_gene_list_bool_array(all_genes_vector, gene_term_map):

    """Represent ontology term gene lists as bool vectors."""

    all_terms = gene_term_map["Term"].drop_duplicates()

    gene_bool_vectors = []
    for term in all_terms:
        genes_in_term = gene_term_map[gene_term_map["Term"] == term]["Gene"]
        gene_bool_vector = all_genes_vector.isin(genes_in_term)
        gene_bool_vector.name = term
        gene_bool_vectors.append(gene_bool_vector)

    gene_list_bool_array = pd.concat(gene_bool_vectors, axis=1)

    return gene_list_bool_array.astype(uint8)


def _get_all_genes_in_term_subtree(data_tuple):
    terms, parent_offspring_map, gene_term_map = data_tuple

    subtree_gene_lists = []
    for term in terms:

        # get all offspring terms
        offspring_mask = parent_offspring_map["Parent"] == term
        offspring = parent_offspring_map["Offspring"][offspring_mask]

        # collect all gene vectors from parent and children in list
        subtree_terms = gene_term_map.columns[
            gene_term_map.columns.isin(offspring)]
        gene_vectors = [gene_term_map[t] for t in subtree_terms]

        # Find genes in at least one term in the tree
        subtree_gene_list = reduce(logical_or, gene_vectors)
        subtree_gene_list.name = term
        subtree_gene_lists.append(subtree_gene_list)

    return pd.concat(subtree_gene_lists, axis=1).astype(uint8)


@MEMORY.cache(verbose=0)
def get_all_genes_in_term_subtree(parent_offspring_map, gene_term_map, nb_cpus):

    """Find all genes associated with all terms below term node.

    Needed because we are not interested merely in genes associated with a term
    gene list, but all genes associated with all terms below the term node."""

    if nb_cpus > 1:
        term_arrays = array_split(gene_term_map.columns.drop_duplicates(), nb_cpus)
        data_tuples = [(terms, parent_offspring_map, gene_term_map) for terms in
                    term_arrays]
        pool = multiprocessing.Pool(processes=nb_cpus)
        term_subtree_dfs = pool.map(_get_all_genes_in_term_subtree, data_tuples)
        all_genes_in_term_subtree = pd.concat(term_subtree_dfs, axis=1)
    else:
        all_genes_in_term_subtree = _get_all_genes_in_term_subtree(
            (gene_term_map.columns.drop_duplicates(), parent_offspring_map,
             gene_term_map))

    assert gene_term_map.shape == all_genes_in_term_subtree.shape
    assert gene_term_map.columns.equals(all_genes_in_term_subtree.columns)
    assert gene_term_map.index.equals(all_genes_in_term_subtree.index)

    return all_genes_in_term_subtree


def test__check_term_genes_in_all_genes_vector(all_genes_vector, caplog):

    _check_term_genes_in_all_genes_vector(all_genes_vector, ["a", "X"])
    assert "not found" in caplog.text()


def _check_term_genes_in_all_genes_vector(all_genes_vector, ontology_genes):

    genes_ontologies = set(ontology_genes)
    genes_universe = set(all_genes_vector)

    nb_genes_not_in_universe = len(genes_ontologies - genes_universe)

    if nb_genes_not_in_universe:
        logging.warning("There are {} number of genes associated with your" \
                        " ontologies that are not found in the universe of" \
                        " genes.".format(nb_genes_not_in_universe))




import logging
import sys
import pandas as pd


from go_overlap.go.terms_to_annotation import get_and_insert_annotations
from go_overlap.compute_set_overlaps.set_overlaps import \
    compute_go_overlap_matrix, compute_non_go_overlaps
from go_overlap.create_term_subtree_gene_matrix.term_subtree_gene_matrix import \
    create_gene_subtree_matrix, create_gene_list_bool_array
from go_overlap.prepare_input_lists.lists_to_bool_vectors import \
    prepare_gene_vectors_for_computation

from biomartian.bm_queries.bm_query import get_attributes, get_datasets, get_marts, get_bm

from go_overlap.statistics.stats import compute_statistics, \
find_appropriate_statistical_test


def get_all_genes(dataset, experiment_genes):

    """Return all gene names from biomart."""

    df = get_bm("external_gene_name", "go_id", dataset, "ensembl")
    df = df["external_gene_name"].drop_duplicates()

    # this is only true if experiment genes was given
    if isinstance(experiment_genes, pd.Series):
        df = df[df.str.upper().isin(experiment_genes.str.upper())]

    return df


def reorder_df(df):

    """Reorder the columns for final output.

    The cols in the df are different depending on which stats method was used
    to compare the difference between input lists.
    """

    if "chi_sq" in df:
        reordered_columns = ["go_root", "ontology", "chi_sq_fdr", "a_hpgm_go_fdr", "a_odds", "b_hpgm_go_fdr",
                             "b_odds", "a_I_b_I_go", "a_M_b_I_go", "b_M_a_I_go", "go_subtree", "U", "a_I_b",
                             "a_M_b", "b_M_a", "chi_sq", "test_obs", "b_hpgm_go", "a_hpgm_go", "go_annotation"]
        df = df.sort("chi_sq_fdr")

    elif "fisher" in df:
        reordered_columns = ["go_root", "ontology", "fisher_fdr", "a_hpgm_go_fdr", "a_odds", "b_hpgm_go_fdr",
                             "b_odds",  "a_I_b_I_go", "a_M_b_I_go", "b_M_a_I_go", "go_subtree", "U", "a_I_b",
                             "a_M_b", "b_M_a", "fisher", "b_hpgm_go", "a_hpgm_go", "go_annotation"]
        df = df.sort("fisher_fdr")

    else:
        reordered_columns = ["go_root", "ontology", "a_hpgm_go_fdr", "a_odds", "a", "a_I_go", "go_subtree",
                             "U", "a_hpgm_go", "go_annotation"]
        df = df.sort("a_hpgm_go_fdr")

    df = df[reordered_columns]

    return df

def _remove_go_terms_with_too_many_genes(df, max_genes_prct_limit):

    if max_genes_prct_limit == 0:
        return df

    go_genes_prct_of_u = df.go_subtree / df.U
    go_terms_with_too_many_genes = go_genes_prct_of_u > max_genes_prct_limit

    df = df.drop(df[go_terms_with_too_many_genes].index)

    return df



def main(ontologies, dataset, gene_vectors, experiment_genes, nb_cpus, max_genes_prct_limit):

    # writes command run as first line to stdout
    print("# " + " ".join(sys.argv))

    all_genes_vector = get_all_genes(dataset, experiment_genes)

    bool_vectors = prepare_gene_vectors_for_computation(gene_vectors,
                                                        all_genes_vector)

    gene_subtree_df = create_gene_subtree_matrix(ontologies, dataset,
                                                 all_genes_vector, nb_cpus)


    go_overlap_matrix = compute_go_overlap_matrix(gene_subtree_df, bool_vectors,
                                                  nb_cpus)

    logging.debug("The overlap data looks like:\n {}".format(go_overlap_matrix))

    go_overlap_matrix = _remove_go_terms_with_too_many_genes(go_overlap_matrix, max_genes_prct_limit)

    df = compute_statistics(go_overlap_matrix, gene_vectors)

    logging.debug("Df with stats:\n {}".format(df))

    df = get_and_insert_annotations(df, dataset, nb_cpus)

    logging.debug("Df with annotations:\n {}".format(df))

    df = reorder_df(df)

    logging.debug("Reordered df:\n {}".format(df))

    df.to_csv(sys.stdout, index=False, header=True, sep="\t", quotechar="'")



import logging


def nb_union_genes_in_go_and_experiment(term_to_genes_map_per_ontology, experiment_gene_list):

    number_genes_union_go_experiment = {}

    experiment_genes = set(experiment_gene_list)

    for ontology, term_to_genes_map in term_to_genes_map_per_ontology.items():
        ontology_genes = set()
        for genes in term_to_genes_map.values():
            ontology_genes.update(genes)

        number_total_genes = len(ontology_genes.union(experiment_genes))
        number_genes_union_go_experiment[ontology] = number_total_genes

        logging.info("Number total genes in the union of the experiment genes and ontology {}: {}".format(ontology, number_total_genes))

    return number_genes_union_go_experiment



from scipy.stats import chisqprob


def lap_ppv(df):

    x1 = df["a_I_b"]
    n1 = x1 + df["a_I_b_I_go"]
    x2 = df["a_M_b_I_go"]
    n2 = x2 + df["a_I_b_I_go"]
    x3 = df["b_M_a_I_go"]
    n3 = df["a_I_b_I_go"] + x3

    nA = n1 + n2
    nB = n1 + n3

    xAG = x1 + x2
    xBG = x1 + x3

    # x1 <- datamat[,5]
    # n1 <- datamat[,1]+datamat[,5]
    # x2 <- datamat[,6]
    # n2 <- datamat[,2]+datamat[,6]
    # x3 <- datamat[,7]
    # n3 <- datamat[,3]+datamat[,7]

    d_prime = (xAG + xBG) / (nA + nB)
    z_prime = nB / (nA + nB)

    numerator = (xBG - nB * (xAG + xBG) / (nA + nB))**2

    denominator = ((0 - d_prime)**2) * ((n1 - x1) * (1 - 2 * z_prime)**2 + (n2 - x2)*(0 - 1 * z_prime)**2+(n3 - x3)*(1 - 1 * z_prime)**2)+((1 - d_prime)**2) * (x1 * (1 - 2 * z_prime)**2 + x2 * (0 - 1 * z_prime)**2 + x3 * (1 - 1 * z_prime)**2)

    test_observation = numerator / denominator

    df["chi_sq"] = chisqprob(test_observation, 1)
    df["test_obs"] = test_observation

    return df



from setuptools import setup

setup(
    name = "goverlap",
    scripts = ["bin/goverlap"],
    packages = ["go_overlap", "go_overlap/compute_set_overlaps",
                "go_overlap/create_term_subtree_gene_matrix",
                "go_overlap/gene_to_terms", "go_overlap/go",
                "go_overlap/go_hierarchy", "go_overlap/prepare_input_lists",
                "go_overlap/statistics", "go_overlap/utils"],
    version = "0.0.2",
    description = "Powerful ontology term enrichment analyzer for the command line",
    author = "Endre Bakken Stovner",
    author_email = "endrebak@stud.ntnu.no",
    url = "http://github.com/endrebak/goverlap",
    keywords = ["Gene Ontology", "Term enrichment"],
    license = ["MIT"],
    install_requires = ["pandas", "docopt", "godb>=0.0.4", "biomartian>=0.0.17",
                        "kg>=0.0.5", "ebs>=0.0.4"],
    classifiers = [
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development :: Libraries :: Python Modules"],
    long_description = ("Powerful ontology term enrichment analyzer for the command line. See the url for more info.")
)



from __future__ import print_function

from numpy import uint8, logical_and, logical_or, array_split, sum as np_sum
from joblib import Memory
import logging
from multiprocessing import Pool
from functools import reduce

import pandas as pd

from go_overlap.gene_to_terms.terms_to_genes import terms_to_genes
from go_overlap.utils.helper_functions import init_worker

MEMORY = Memory(cachedir="./")


@MEMORY.cache(verbose=0)
def compute_go_overlap_matrix(subtree_gene_list, gene_vectors,
                              nb_cpus, experiment_genes=None):

    """Find overlap between input genes and ontology term subtree genes.

    First gets the ontology term to gene maps needed, then computes overlap.

    The final dataframe contains the following columns:

    - go_root: root of the subtree we are computing overlap counts for
    - ontology: ontology "go_root" belongs to
    - a_I_b_I_go: nb genes in a and b and ontology
    - a_M_b_I_go: nb genes in a (but not b) and in ontology
    - b_M_a_I_go: nb genes in b (but not a) and in ontology
    - go_tree: nb genes in subtree of go_root
    - U: nb genes in universe (union of experiment genes and all genes in ontology)
    - a_I_b: nb genes in a and b
    - a_M_b: nb genes in a but not b
    - b_M_a: nb genes in b but not a

    These are used to compute statistics further down the pipeline."""

    #
    logging.info("Find number of genes in both input gene lists and term genes.")
    subtree_overlaps = compute_subtree_overlaps(subtree_gene_list,
                                                gene_vectors, nb_cpus)

    logging.info("Find number of genes in input gene list.")
    non_go_intersections = create_df_from_gene_vectors(gene_vectors,
                                                       subtree_gene_list.columns)
    logging.info("Compute number of genes in universe.")
    genes_in_universe = compute_size_universe(subtree_gene_list,
                                              experiment_genes)
    nb_go_subtree_genes = find_nb_go_subtree_genes(subtree_gene_list)

    logging.info("Concatenating overlap data.")
    df = pd.concat([subtree_overlaps, non_go_intersections, genes_in_universe,
                      nb_go_subtree_genes], axis=1)

    # Making the go term index into a data column
    df = df.reset_index()
    df.rename(columns={"index": "go_root"}, inplace=True)

    return df


def create_df_from_gene_vectors(gene_vectors, indices):
    """Create df of number of genes in input gene lists.

    Each value in each column should be the same."""

    non_go_intersections = pd.concat(gene_vectors, axis=1)
    non_go_intersections = non_go_intersections.sum()
    non_go_intersections = pd.concat([non_go_intersections] * len(gene_vectors),
                                     axis=1)
    non_go_intersections = [non_go_intersections.T.head(1)] * len(indices)
    non_go_intersections = pd.concat(non_go_intersections)

    non_go_intersections = non_go_intersections.set_index(indices)

    return non_go_intersections


def find_nb_go_subtree_genes(subtree_gene_lists):

    """Find the number of genes in each ontology subtree."""

    go_subtree = subtree_gene_lists.sum()
    go_subtree.name = "go_subtree"

    return go_subtree


def compute_size_universe(subtree_gene_lists, experiment_genes):

    """Find number of genes that are either in the experiment or ontologies.

    Use just the total number of genes existing (annotated), if no experiment
    genes given.

    Including the experiment genes is a way to make the p-values less
    conservative."""

    if experiment_genes:
        all_genes_in_ontologies = reduce(subtree_gene_lists, logical_or)
        genes_in_universe = logical_or(all_genes_in_ontologies, experiment_genes)
        nb_genes_in_universe = sum(genes_in_universe)
    else:
        nb_genes_in_universe = len(subtree_gene_lists)

    nb_genes_in_universe = pd.Series(nb_genes_in_universe,
                                     subtree_gene_lists.columns)
    nb_genes_in_universe.name = "U"

    return nb_genes_in_universe


def compute_subtree_overlaps(subtree_gene_list, gene_vectors, nb_cpus):

    """Compute overlap between go term genes and lists of genes.

    TODO: rename internal "subtree_gene_list"; overwrites outer"""

    subtree_gene_list, gene_vectors = _remove_genes_that_cannot_matter(subtree_gene_list,
                                                                       gene_vectors)
    intersected_gene_lists = []
    subtree_gene_lists = array_split(subtree_gene_list, nb_cpus, axis=1)

    pool = Pool(processes=nb_cpus)

    for gene_list in gene_vectors:

        data_tuples = [(gene_list, subtree_gene_list) for subtree_gene_list in
                       subtree_gene_lists]

        intersected_gene_dfs = pool.map(_compute_go_intersections, data_tuples)
        intersected_gene_df = pd.concat(intersected_gene_dfs)

        intersected_gene_lists.append(intersected_gene_df)

    overlap_df = pd.concat(intersected_gene_lists, axis=1)

    return overlap_df

def _remove_genes_that_cannot_matter(go_term_genes, gene_vectors):

    # Remove all genes not in a or b, because they cannot be in the intersection
    # of a and GO (or b and go)

    # this would typically reduce the matrix used from U rows to (a union b) genes
    # a reduction from perhaps 25k to 1k or less, which means time and memory saved

    genes_that_dont_matter = _find_the_genes_that_dont_matter(gene_vectors)

    gene_vectors = [g.drop(genes_that_dont_matter) for g in gene_vectors]
    go_term_genes = go_term_genes.drop(genes_that_dont_matter)

    return go_term_genes, gene_vectors

def _find_the_genes_that_dont_matter(gene_vectors):


    genes_in_at_least_one_list = reduce(logical_or, gene_vectors)
    genes_that_dont_matter = genes_in_at_least_one_list[genes_in_at_least_one_list == False].index

    return genes_that_dont_matter



def _compute_go_intersections(gene_vector_term_genes_array):

    """Compute overlap between go term genes and a list of genes."""

    gene_vector, term_genes_array = gene_vector_term_genes_array
    result_df = logical_and(term_genes_array, gene_vector[:, None])
    result_df.name = gene_vector.name
    result_series = result_df.sum()
    result_series.name = gene_vector.name + "_I_go"

    return result_series


def compute_non_go_overlaps(gene_vectors):

    """Compute overlap intersection and difference between input gene lists."""

    # if only one vector there is no overlap to compute so short circuit
    gene_vectors[0].name = "a"
    if len(gene_vectors) == 1:
        return gene_vectors

    gene_vectors[1].name = "b"

    a, b = gene_vectors
    a_I_b = logical_or(a, b)
    a_I_b.name = "a_I_b"

    a_M_b = a & ~b
    a_M_b.name = "a_M_b"

    b_M_a = b & ~a
    b_M_a.name = "b_M_a"

    return [a, b, a_I_b, a_M_b, b_M_a]
